/*
    This sample illustrates how to get and set camera parameter values.

    For camera configuration and access to other parameters, the GenICam GenApi
    library is used. The GenICam GenApi specification
    (http://www.emva.org/standards-technology/genicam/) defines a format for
    camera description files.
    These files describe the configuration interface of GenICam-compliant
    cameras. The description files are written in XML and describe camera
    registers, their interdependencies, and all other information required for
    accessing high-level features such as Exposure Time, Confidence Threshold,
    or Pixel Format by means of low-level register read and write operations.

    The GenApi library represents the elements of a camera description file
    as software objects called "nodes". For example, a node can represent a
    single camera register, a camera parameter such as Confidence Threshold, a
    set of available parameter values, etc. Each node implements the
    GenApi::INode interface.

    The nodes are linked by different relationships as explained in the
    GenApi standard document available at www.GenICam.org. The complete set of
    nodes is stored in a data structure called a "node map".
    At runtime, a node map is instantiated from an XML description file 
    downloaded from the camera.

    The names and types of the parameter nodes can be found by using the blaze 
    Viewer tool.

    The CBlazeCamera::GetParameter() function is used for retrieving smart
    pointers providing access to the parameter nodes.

    The GenICam GenApi standard document can be downloaded from 
    http://www.emva.org/wp-content/uploads/GenICam_Standard_v2_0.pdf
*/


#include <ConsumerImplHelper/BlazeCamera.h>
#include <string>
#include <iostream>

using namespace GenTLConsumerImplHelper;
using namespace GenApi;


// Adjust a value to make it comply with range and increment constraints.
//
// The parameter's minimum and maximum are always considered valid values.
// If the increment is larger than one, the returned value will be: min + (n * inc).
// If the value doesn't meet these criteria, it will be rounded down to ensure compliance.
int64_t Adjust(int64_t val, const CIntegerPtr ptrInt)
{
    const int64_t minimum = ptrInt->GetMin();
    const int64_t maximum = ptrInt->GetMax();
    const int64_t inc = ptrInt->GetInc();
    
    // Check the lower bound.
    if (val < minimum)
    {
        return minimum;
    }

    // Check the upper bound.
    if (val > maximum)
    {
        return maximum;
    }

    // Check the increment.
    if (inc == 1)
    {
        // Special case: all values are valid.
        return val;
    }
    else
    {
        // The value must be min + (n * inc).
        // Due to the integer division, the value will be rounded down.
        return minimum + ( ((val - minimum) / inc) * inc );
    }
}

int main(int argc, char* argv[])
{
    int exitCode = EXIT_FAILURE;

    CBlazeCamera camera;
    try
    {
        CBlazeCamera::InitProducer();
        camera.OpenFirstCamera();

        // 
        // Common properties of all parameter types
        //

        // We use GenApi's smart pointer like classes for a convenient way to get access 
        // to parameters. GenApi supports different parameter types, for each of them 
        // there is a corresponding parameter class: CIntgegerPtr, CFloatPtr, CEnumerationPtr, 
        // CBooleanPtr, etc.
        CIntegerPtr ptrFancyParameter = camera.GetParameter( "FancyParameter" );

        // Although the camera doesn't provide this fancy parameter, the camera class
        // returns a smart pointer. Accessing the smart pointer will throw an exception 
        // since there is no corresponding GenApi node object the pointer could provide 
        // access to. Trying to access the value or the properties of ptrFancyParameter will result
        // in an GenICam::GenericException. 
        // Check to see whether GetParameter returned a valid pointer to a parameter:
        if ( ! ptrFancyParameter.IsValid() )
        {
            std::cout << "Camera doesn't provide the parameter 'FancyParameter'" << std::endl;
        }

        // Parameters might be read-only. When accessing read-only parameters, we will get
        // an exception. The GenApi::IsWritable() function can be used to check whether
        // a value can be set.
        // GenApi provides some further convenience functions for checking the accessibility
        // of nodes: GenApi::IsReadable(), GenApi::IsImplemended().
        // These functions apply to all parameter types.

        // Of course, a non-existent parameter is not readable...
        if ( ! IsReadable(ptrFancyParameter) )
        {
            std::cout << "'FancyParameter' is not readable."  << std::endl;
        }

        // Measured temperatures are not writable...
        CFloatPtr ptrDeviceTemperature = camera.GetParameter ( "DeviceTemperature" );
        if ( ! IsWritable(ptrDeviceTemperature) )
        {
            std::cout << "Device temperature is not writable." << std::endl;
        }
        if (IsReadable(ptrDeviceTemperature))
        {
            std::cout << "Current temperature is " << ptrDeviceTemperature->GetValue() << " degrees Celsius" << std::endl;
        }

        //
        // Usage of integer parameters
        //

        // Some properties may have restrictions, i.e., the values must be a multiple of a certain
        // increment. Use GetInc/GetMin/GetMax to make sure you set a valid value, as shown
        // in the Adjust() function above.
        GenApi::CIntegerPtr ptrConfidenceThreshold = camera.GetParameter("ConfidenceThreshold");
        // As this is for demonstration purposes only, we don't change the value.
        int64_t newThresh = ptrConfidenceThreshold->GetValue(); 
        newThresh = Adjust(newThresh, ptrConfidenceThreshold);
        ptrConfidenceThreshold->SetValue(newThresh);

        // Values of all parameter types can be converted to a string using the ToString() method.
        std::cout << "Confidence Threshold : " << ptrConfidenceThreshold->ToString() << std::endl;

        // 
        // Usage of enumeration parameters, selectors, and Boolean parameters
        //
        // Enumeration parameters, as the PixelFormat, represent a defined set of named values.
        // Selectors are enumeration parameters influencing the state of the parameters
        // that are "selected by" the selector.
        // The ComponentSelector feature, as an example, is a selector. The feature controls 
        // the number of parts an acquired image consists of. Each part can be individually 
        // enabled by using the ComponentEnable parameter.
        // For some parts there are multiple options for the part's pixel format.
        // Thus, the ComponentSelector "selects" the PixelFormat and the Component parameters.
        //
        // All parameter types provide a corresponding FromString() method allowing to set
        // a parameter by using a string. The FromString() method is the only way to set the
        // value of an enumeration parameter.
        // 
        // Example: Iterate over all parts and enable those parts that support RGB data.

        CEnumerationPtr ptrComponentSelector = camera.GetParameter("ComponentSelector");
        CBooleanPtr ptrComponentEnable = camera.GetParameter("ComponentEnable");
        CEnumerationPtr ptrPixelFormat( camera.GetParameter("PixelFormat") );
        NodeList_t entries;
        ptrComponentSelector->GetEntries( entries );
        for ( NodeList_t::const_iterator entry = entries.begin(); entry != entries.end(); ++entry)
        {
            CEnumEntryPtr ptrEntry( *entry );
            std::cout <<  ptrEntry->GetSymbolic() << " part is " << (IsAvailable( ptrEntry ) ? "supported" :  "not supported") << std::endl;
            if ( IsAvailable( ptrEntry ) )
            {
                // First, select the part to configure...
                ptrComponentSelector->FromString( ptrEntry->GetSymbolic() );
                // ... then access the parameters for the part.
                if ( IsAvailable( ptrPixelFormat->GetEntryByName( "RGB8")))
                {
                    ptrPixelFormat->FromString( "RGB8");
                    ptrComponentEnable->SetValue(true);
                }
                std::cout << "Pixel format of part " << ptrComponentSelector->ToString() << ": " << ptrPixelFormat->ToString() << std::endl;
                std::cout << "Part is " << (ptrComponentEnable->GetValue() ? "enabled" : "disabled") << std::endl;
            }
        }

        //
        // Usage of floating point parameters
        //
        CFloatPtr ptrAcquisitionFrameRate( camera.GetParameter("AcquisitionFrameRate") );
        std::cout << "Current acquisition frame rate: " << ptrAcquisitionFrameRate->GetValue();
        // Set to maximum
        ptrAcquisitionFrameRate->SetValue( ptrAcquisitionFrameRate->GetMax() );
        std::cout << ". New value: " << ptrAcquisitionFrameRate->GetValue() << std::endl;

        //
        // Configuring the camera's exposure time
        // Note: Generally, you should not adjust the exposure time. Unless you
        // don't notice saturation effects or missing depth information due
        // to close objects being overexposed, you should stick to the default
        // exposure time value. 
        //
        CFloatPtr ptrExposureTime(camera.GetParameter("ExposureTime"));
        // Retrieve the current exposure time.
        double currentExposureTime = ptrExposureTime->GetValue();
        // For demonstration purposes, we write back the current value.
        ptrExposureTime->SetValue(currentExposureTime);
        std::cout << "Set exposure time to " << ptrExposureTime->GetValue() << std::endl;

        camera.Close();
        exitCode = EXIT_SUCCESS;
    }
    catch ( GenICam::GenericException& e )
    {
        std::cerr << "Exception occurred: " << std::endl << e.GetDescription() << std::endl;
        // After successfully opening the camera, the IsConnected method can be used 
        // to check if the device is still connected.
        if ( camera.IsOpen() && ! camera.IsConnected() )
        {
            std::cerr << "Camera has been removed." << std::endl;
        }
    }

    // Release the GenTL producer and all of its resources. 
    // Note: Don't call TerminateProducer() until the destructor of the CBlazeCamera
    // class has been called. The destructor may require resources which may not
    // be available anymore after TerminateProducer() has been called.
    if ( CBlazeCamera::IsProducerInitialized() )
        CBlazeCamera::TerminateProducer();  // Won't throw any exceptions

    std::cout << std::endl << "Press Enter to exit." << std::endl;
    while (std::cin.get() != '\n');

    return exitCode;
}

