/*
This sample illustrates how to grab images from a blaze camera and how to access 
the depth data.

Further Reading:

The 'CameraConfiguration' sample illustrates in more detail how to configure a blaze camera.
The 'SimpleGrabLoop' sample illustrates how to use a class member function as grab callback.
The 'CustomGrabLoop' sample illustrates more advanced use cases like how to use user-provided buffers
and how an application can provide its own grab loop.

*/

#include <iostream>
#include <iomanip> 
#include <atomic>
#include <ConsumerImplHelper/BlazeCamera.h>

using namespace GenTLConsumerImplHelper;

int main(int argc, char* argv[])
{
    int exitCode = EXIT_SUCCESS;

    try
    {
        CBlazeCamera::InitProducer();

        CBlazeCamera camera;

        // Open the first camera found, i.e., establish a connection to the camera device. 
        camera.OpenFirstCamera();
        std::cout << "Connected to camera " << camera.GetCameraInfo().strDisplayName << std::endl;

        // Enable 3D (point cloud) data
        GenApi::CEnumerationPtr ptrComponentSelector = camera.GetParameter("ComponentSelector");
        ptrComponentSelector->FromString("Range");
        GenApi::CBooleanPtr ptrComponentEnable = camera.GetParameter("ComponentEnable");
        ptrComponentEnable->SetValue(true);
        // Range information can be sent either as a 16-bit grey value image or as 3D coordinates (point cloud). 
        // For this sample, we want to acquire 3D coordinates.
        // Note: To change the format of an image component, the Component Selector must first be set to the component
        // you want to configure (see above).
        // To use 16-bit integer depth information, choose "Coord3D_C16" instead of "Coord3D_ABC32f". For more details 
        // about how to process 16-bit depth images, refer to the GrabDepthMap sample.
        GenApi::CEnumerationPtr ptrPixelFormat = camera.GetParameter("PixelFormat");
        ptrPixelFormat->FromString("Coord3D_ABC32f");

        // Data to be shared with the the grab callback
        struct Context
        {
            std::atomic<int> nBuffersGrabbed{0};
        } context;
        auto sharedContext = std::ref(context); // holds a reference to context

        // Acquire images until the lambda function used as grab callback indicates to stop acquisition. 
        // 5 buffers are used (round-robin). The grab timeout is set to 1000 ms
        // Note that since the lambda function will be called by a different thread, it has to 
        // capture variables by value only. Capturing by reference is undefined behavior!
        camera.GrabContinuous(5, 1000, [sharedContext](GrabResult grabResult, BufferParts parts) -> bool {
            // Note that sharedContext is captured by value. Since it holds a reference, 
            // we can use sharedContext to access shared data, though.
            if (grabResult.status == GrabResult::Timeout)
            {
                std::cerr << "Timeout occurred. Acquisition stopped." << std::endl;
                return false; // Indicate to stop acquisition
            }
            const int nBuffersGrabbed = ++sharedContext.get().nBuffersGrabbed;
            if (grabResult.status != GrabResult::Ok)
            {
                std::cerr << "Image " << nBuffersGrabbed << " was not grabbed successfully." << std::endl;
            }
            else
            {
                // Retrieve the values for the center pixel. 
                const int x = (int)(0.5 * parts[0].width); 
                const int y = (int)(0.5 * parts[0].height);
                auto p3DCoordinate = (CBlazeCamera::Coord3D*) parts[0].pData + x + y * parts[0].width;
                std::cout.setf(std::ios_base::fixed);
                std::cout.precision(1);
                if (p3DCoordinate->z != 0)
                    std::cout << "x=" << std::setw(6) << p3DCoordinate->x << " y=" << std::setw(6) << p3DCoordinate->y << " z=" << std::setw(6) << p3DCoordinate->z << "\n";
                else
                    std::cout << "x=   n/a y=   n/a z=   n/a\n";
            }
            return nBuffersGrabbed < 10; // Indicate to stop acquisition when 10 buffers are grabbed
            });

        // GrabContinuous() blocks execution until the grab callback returns false.

        // Clean-up
        camera.Close();
    }
    catch (GenICam::GenericException& e)
    {
        std::cerr << "Exception occurred: " << std::endl << e.GetDescription() << std::endl;
        exitCode = EXIT_FAILURE;
    }

    // Release the GenTL producer and all of its resources. 
    // Note: Don't call TerminateProducer() until the destructor of the CBlazeCamera
    // class has been called. The destructor may require resources which may not
    // be available anymore after TerminateProducer() has been called.
    if (CBlazeCamera::IsProducerInitialized())
        CBlazeCamera::TerminateProducer();  // Won't throw any exceptions

    std::cout << std::endl << "Press Enter to exit." << std::endl;
    while (std::cin.get() != '\n');

    return exitCode;
}
